/*
 * 
 */
package casa.JION.console;

import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.rmi.MarshalledObject;

import casa.JION.entry.Entry;
import casa.JION.event.RemoteEventListener;
import casa.JION.lease.Lease;
import casa.JION.space.JavaSpace;
import casa.JION.space.JavaSpaceSingleton;
import casa.dodwan.docware.Descriptor;
import casa.dodwan.util.ArgumentParsing;

// TODO: Auto-generated Javadoc
/**
 * The Class Console.
 */
public class Console extends casa.dodwan.util.Console {
	
	/** The js. */
	private JavaSpace js=null;
	
	/**
	 * Instantiates a new console.
	 */
	public Console(){
		js=JavaSpaceSingleton.getInstance();
	}
	
	/* (non-Javadoc)
	 * @see casa.dodwan.util.Console#help(java.lang.String, java.io.PrintStream)
	 */
	/** {@inheritDoc} */
	@Override
	public void help(String offset, PrintStream out){
		out.println(
				"JavaSpaces command set:\n"
				+offset+"[write]  [-e entry]      [-d duration]\n"
				+offset+"[read]   [-tmpl template][-t timeout]\n"
				+offset+"[take]   [-tmpl template][-t timeout]\n"
				+offset+"[notify] [-tmpl template][-t timeout]\n"
				+offset+"[readif] [-tmpl template]\n"				
				+offset+"[takeif] [-tmpl template]\n"
				+offset+"[renew]  [-id lease id]  [-d duration]\n"
				+offset+"[cancel] [-id lease id]\n"
				+offset+"[detail] [-id entry id]\n"					
				+offset+"[list]  entries  id/lease\n"
				+offset+"[list2] notifies id/lease\n"
				+offset+"[h]elp");
	}
	
	/* (non-Javadoc)
	 * @see casa.dodwan.util.Console#processCommand(java.lang.String[], java.io.BufferedReader, java.io.PrintStream)
	 */
	/** {@inheritDoc} */
	@Override
	public void processCommand(String[] arguments, BufferedReader reader,
		       PrintStream out) {
		boolean ok=false;
		if (arguments.length <1) {
			ok=true;
			help("  ",out);
			return;
		}
		
		if (arguments[0].startsWith("notify")){
			ok=true;
			notify(arguments,reader,out);
			return;
		}
		
		if (arguments[0].startsWith("readif")){
			ok=true;
			readif(arguments,reader,out);
			return;
		}
		
		if (arguments[0].startsWith("read")){
			ok=true;
			read(arguments,reader,out);
			return;
		}
		
		if (arguments[0].startsWith("takeif")){
			ok=true;
			takeif(arguments,reader,out);
			return;
		}
		
		if (arguments[0].startsWith("take")){
			ok=true;
			take(arguments,reader,out);
			return;
		}
		
		if (arguments[0].startsWith("write")){
			ok=true;
			write(arguments,reader,out);
			return;
		} 
		
		if (arguments[0].startsWith("list2")){
			ok=true;
			list2(arguments,reader,out);
			return;
		} 
		
		if (arguments[0].startsWith("list")){
			ok=true;
			list(arguments,reader,out);
			return;
		} 
		
		if (arguments[0].startsWith("detail")){
			ok=true;
			detail(arguments,reader,out);
			return;
		} 
		
		if (arguments[0].startsWith("renew")){
			ok=true;
			renew(arguments,reader,out);
			return;
		}
		
		if (arguments[0].startsWith("cancel")){
			ok=true;
			cancel(arguments,reader,out);
			return;
		}
		
		if (arguments[0].startsWith("h")){
			ok=true;
			help("   ", out);
			return;
		}
		
		if (!ok){
			out.println("  Wrong command, please use [h]elp for more informations");
			return;
		}
	}
	
	
	/**
	 * List2.
	 *
	 * @param arguments the arguments
	 * @param reader the reader
	 * @param out the out
	 */
	private void list2(String[] arguments, BufferedReader reader, PrintStream out){
		js.listNotifies(out);
	}
	
	/**
	 * List.
	 *
	 * @param arguments the arguments
	 * @param reader the reader
	 * @param out the out
	 */
	private void list(String[] arguments, BufferedReader reader, PrintStream out){
		js.listEntries(out);
	}
	
	/**
	 * Detail.
	 *
	 * @param arguments the arguments
	 * @param reader the reader
	 * @param out the out
	 */
	private void detail(String[] arguments, BufferedReader reader, PrintStream out){
		String id=null;
		try {
			id=ArgumentParsing.searchString("-id", arguments);
		} catch (Exception e) {
			out.println(e.getMessage());
			e.printStackTrace();
		}
		if(id==null){
			out.println("The entry id can't be null");
			return;
		}
		out.println(js.detail(id));
	}
	
	/**
	 * Readif.
	 *
	 * @param arguments the arguments
	 * @param reader the reader
	 * @param out the out
	 */
	private void readif(String[] arguments, BufferedReader reader, PrintStream out){
		String tmpl=null;
		try {
						tmpl= ArgumentParsing.searchString("-tmpl", arguments);
		} catch (Exception e) {
			out.println(e.getMessage());
			e.printStackTrace();
		}
		if(tmpl==null){
			out.println("The template or the timeout can't be null");
			return;
		}
		
		Entry template=new Entry();
		template.setDescriptor(stringToDescriptor(tmpl));
		try {
			out.println(js.read(template, null, net.jini.space.JavaSpace.NO_WAIT));
		} catch (Exception e) {
			out.println(e.getMessage());
			e.printStackTrace();
		}
	}
	
	/**
	 * Cancel.
	 *
	 * @param arguments the arguments
	 * @param reader the reader
	 * @param out the out
	 */
	private void cancel(String[] arguments, BufferedReader reader, PrintStream out){
		String id=null;
		try {
			id=ArgumentParsing.searchString("-id", arguments);
		} catch (Exception e) {
			out.println(e.getMessage());
			e.printStackTrace();
		}
		if(id==null){
			out.println("The lease id can't be null");
			return;
		}
		try {
			Lease lease=js.getLease(id);
			lease.cancel();
			out.println("The lease: "+id+" has been cancelled");
		} catch (Exception e) {
			out.println(e.getMessage());
			e.printStackTrace();
		}	
	}
	
	/**
	 * Renew.
	 *
	 * @param arguments the arguments
	 * @param reader the reader
	 * @param out the out
	 */
	private void renew(String[] arguments, BufferedReader reader, PrintStream out){
		String id=null;
		String duration=null;
		try {
			id=ArgumentParsing.searchString("-id", arguments);
			duration=ArgumentParsing.searchString("-d", arguments);
		} catch (Exception e) {
			out.println(e.getMessage());
			e.printStackTrace();
		}
		if(id==null||duration==null){
			out.println("The lease id or the duration can't be null");
			return;
		}
		try {
			Lease lease=js.getLease(id);
			lease.renew(Long.valueOf(duration).longValue());
			out.println("The lease: "+id+" has been renewed for an additional period of time");
		} catch (Exception e) {
			out.println(e.getMessage());
			e.printStackTrace();
		}
	}

	
	/**
	 * Write.
	 *
	 * @param arguments the arguments
	 * @param reader the reader
	 * @param out the out
	 */
	private void write(String[] arguments, BufferedReader reader, PrintStream out){
		String timeout=null;
		String e=null;
		try {
			timeout=ArgumentParsing.searchString("-d", arguments);
			e= ArgumentParsing.searchString("-e", arguments);
		} catch (Exception ex) {
			out.println(ex.getMessage());
			ex.printStackTrace();
		}
		if(timeout==null||e==null){
			out.println("The entry or the duration can't be null");
			return;
		}
		
		Entry entry=new Entry();
		entry.setDescriptor(stringToDescriptor(e));
		try {
			js.write(entry, null, Long.valueOf(timeout).longValue());
			out.println("The entry has been added to the space");
		} catch (Exception e1) {
			out.println(e1.getMessage());
			e1.printStackTrace();
		}
	}
	
	/**
	 * Takeif.
	 *
	 * @param arguments the arguments
	 * @param reader the reader
	 * @param out the out
	 */
	private void takeif(String[] arguments, BufferedReader reader, PrintStream out){
		String tmpl=null;
		try {
			tmpl= ArgumentParsing.searchString("-tmpl", arguments);
		} catch (Exception e) {
			out.println(e.getMessage());
			e.printStackTrace();
		}
		if(tmpl==null){
			out.println("The template can't be null");
			return;
		}
		Entry template=new Entry();
		template.setDescriptor(stringToDescriptor(tmpl));
		try {
			out.println(js.take(template, null, net.jini.space.JavaSpace.NO_WAIT));
		} catch (Exception e) {
			out.println(e.getMessage());
			e.printStackTrace();
		}
	}
	
	/**
	 * Take.
	 *
	 * @param arguments the arguments
	 * @param reader the reader
	 * @param out the out
	 */
	private void take(String[] arguments, BufferedReader reader, PrintStream out){
		String timeout=null;
		String tmpl=null;
		try {
			timeout=ArgumentParsing.searchString("-t", arguments);
			tmpl= ArgumentParsing.searchString("-tmpl", arguments);
		} catch (Exception e) {
			out.println(e.getMessage());
			e.printStackTrace();
		}
		if(timeout==null||tmpl==null){
			out.println("The template or the timeout can't be null");
			return;
		}
		Entry template=new Entry();
		template.setDescriptor(stringToDescriptor(tmpl));
		try {
			out.println(js.take(template, null, Long.valueOf(timeout).longValue()));
		} catch (Exception e) {
			out.println(e.getMessage());
			e.printStackTrace();
		}
	}
	
	
	/**
	 * Notify.
	 *
	 * @param arguments the arguments
	 * @param reader the reader
	 * @param out the out
	 */
	private void notify(String[] arguments, BufferedReader reader, PrintStream out){
		String timeout=null;
		String tmpl=null;
		try {
			timeout=ArgumentParsing.searchString("-t", arguments);
			tmpl= ArgumentParsing.searchString("-tmpl", arguments);
		} catch (Exception e) {
			out.println(e.getMessage());
			e.printStackTrace();
		}
		if(timeout==null||tmpl==null){
			out.println("The template or the timeout can't be null");
			return;
		}
		Entry template=new Entry();
		template.setDescriptor(stringToDescriptor(tmpl));
		try {
			out.println(js.notify(template, null,new RemoteEventListener(), Long.valueOf(timeout).longValue(),
					new MarshalledObject<Object>(null)));
		} catch (Exception e) {
			out.println(e.getMessage());
			e.printStackTrace();
		}
	}
	
	/**
	 * Read.
	 *
	 * @param arguments the arguments
	 * @param reader the reader
	 * @param out the out
	 */
	private void read(String[] arguments, BufferedReader reader, PrintStream out){
		String timeout=null;
		String tmpl=null;
		try {
			timeout=ArgumentParsing.searchString("-t", arguments);
			tmpl= ArgumentParsing.searchString("-tmpl", arguments);
		} catch (Exception e) {
			out.println(e.getMessage());
			e.printStackTrace();
		}
		if(timeout==null||tmpl==null){
			out.println("The template or the timeout can't be null");
			return;
		}
		
		Entry template=new Entry();
		template.setDescriptor(stringToDescriptor(tmpl));
		try {
			out.println(js.read(template, null, Long.valueOf(timeout).longValue()));
		} catch (Exception e) {
			out.println(e.getMessage());
			e.printStackTrace();
		}
		
	}
	
	/**
	 * String to descriptor.
	 *
	 * @param str the str
	 * @return the descriptor
	 * @throws IllegalArgumentException the illegal argument exception
	 */
	private Descriptor stringToDescriptor(String str) throws IllegalArgumentException{
		Descriptor desc=new Descriptor();
		String[] patterns=str.split(",");
		for (int i =0; i< patterns.length;i++) {
			String[] tpattern=patterns[i].split("=");
			if(tpattern.length!=2)
				throw new IllegalArgumentException("Selector format is wrong. It must be attribut=value,attribut=value...");
			desc.setAttribute(tpattern[0], (tpattern[1].equalsIgnoreCase("null"))?".*":tpattern[1]);
		}
//		Descriptor temp=new Descriptor();
//		temp.setAttribute("name", "hala");
//		temp.setAttribute("age", "28");
//		System.out.println(temp.matches(desc));
		return desc;
	}
	
	/**
	 * The main method.
	 *
	 * @param args the arguments
	 */
	public static void main (String[] args){
		Boolean x=null;
		System.out.println((x==null)?".*":x);
		new Console().processCommand(args, new BufferedReader(new InputStreamReader(System.in)), System.out);
	}
}
