/*
 * 
 */
package casa.JION.event;

import net.jini.core.lease.Lease;

/**
 * A utility class for use as a return value for event-interest registration
 * methods.  Objects of this class are meant to encapsulate the information
 * needed by a client to identify a notification as a response to a
 * registration request and to maintain that registration request.  It is
 * not mandatory for an event-interest registration method to use this class.
 * <p>
 * A registration of interest in some kind of event that occurs within the
 * scope of a transaction is leased in the same way as other event interest
 * registrations.  However, the duration of the registration is the minimum
 * of the length of the lease and the duration of the transaction.  Simply
 * put, when the transaction ends (either because of a commit or an abort)
 * the interest registration also ends.  This is true even if the lease for
 * the event registration has not expired and no call has been made to
 * cancel the lease.
 *
 * @author Sun Microsystems, Inc.
 *
 * @since 1.0
 */
public class EventRegistration extends net.jini.core.event.EventRegistration{

	/** The Constant serialVersionUID. */
	private static final long serialVersionUID = 8142606772964345719L;

    /**
     * Constructs an <tt>EventRegistration</tt> object.
     *
     * @param eventID  a <tt>long</tt> representing the event identifier
     * @param source   an <tt>Object</tt> representing the event source
     * @param lease    the registration <tt>Lease</tt> object
     * @param seqNum   a <tt>long</tt> representing the current
     *                 sequence number
     */
	public EventRegistration(long eventID, Object source, Lease lease,
			long seqNum) {
		super(eventID, source, lease, seqNum);
	}
	

	@Override
	public String toString() {
		StringBuffer buffer=new StringBuffer();
		casa.JION.lease.Lease myLease = (casa.JION.lease.Lease) lease; 
		buffer.append("The event source: "+source.toString()+"\n");
		buffer.append("The event id: "+eventID+"\n");
		buffer.append("The event sequence: "+seqNum+"\n");
		buffer.append("The event lease: "+myLease+"\n");
		return buffer.toString();
	}

}
