/*
 * 
 */
package casa.JION.lease;

import java.rmi.RemoteException;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import net.jini.core.lease.LeaseMapException;


/**
 * An unsynchronized Map from Lease to Long (the duration to use when
 * renewing the lease).  For all methods of LeaseMap except canContainKey,
 * an IllegalArgumentException is thrown if a key is not a Lease or a value
 * is not a Long.  Null keys and values are not supported.
 *
 * @author Sun Microsystems, Inc.
 *
 * @since 1.0
 */
public class LeaseMap implements net.jini.core.lease.LeaseMap {

	/** The map. */
	protected HashMap<Lease, Long> map=null;
	
	/** The exception map. */
	protected HashMap<Lease, Exception> exceptionMap=null;
	
	/**
	 * Instantiates a new lease map.
	 *
	 * @param lease the lease
	 * @param duration the duration
	 */
	public LeaseMap(Lease lease, long duration){
		map = new HashMap<Lease, Long>();
		exceptionMap = new HashMap<Lease, Exception>();
		put(lease, duration);
	}
	
	/**
	 * Instantiates a new lease map.
	 *
	 * @param map the map
	 * @param lease the lease
	 * @param duration the duration
	 */
	public LeaseMap(Map<Lease, Long> map, Lease lease, long duration){
		this.map=(HashMap<Lease, Long>)map;
		exceptionMap = new HashMap<Lease, Exception>();
		put(lease, duration);
	}
	
	/**
	 * Check key.
	 *
	 * @param key the key
	 * @return true, if successful
	 */
	protected boolean checkKey(Object key){
		if (key instanceof net.jini.core.lease.Lease)
			return true;
		return false;
	}
	
	/**
	 * Check value.
	 *
	 * @param value the value
	 * @return true, if successful
	 */
	protected boolean checkValue(Object value){
		if (value instanceof Long)
			return true;
		return false;
	}
	
    /**
     * Returns the number of key-value mappings in this map.
     *
     * @return the number of key-value mappings in this map
     */
	public int size() {
		return map.size();
	}

    /**
     * Returns <tt>true</tt> if this map contains no key-value mappings.
     *
     * @return <tt>true</tt> if this map contains no key-value mappings
     */
	public boolean isEmpty() {
		return map.isEmpty();
	}

    /**
     * Returns <tt>true</tt> if this map contains a mapping for the
     * specified key.
     *
     * @param   key   The key whose presence in this map is to be tested
     * @return <tt>true</tt> if this map contains a mapping for the specified
     * key.
     */
	public boolean containsKey(Object key) {
		if (!checkKey(key))
			throw new IllegalArgumentException("Key is not a Lease object");
		return map.containsKey(key);
	}

    /**
     * Returns <tt>true</tt> if this map maps one or more keys to the
     * specified value.
     *
     * @param value value whose presence in this map is to be tested
     * @return <tt>true</tt> if this map maps one or more keys to the
     *         specified value
     */
	public boolean containsValue(Object value) {
		if (!checkValue(value))
			throw new IllegalArgumentException("Value is not a Long object");
		return map.containsValue(value);
	}

    /**
     * Returns the value to which the specified key is mapped,
     * or {@code null} if this map contains no mapping for the key.
     *
     * <p>More formally, if this map contains a mapping from a key
     * {@code k} to a value {@code v} such that {@code (key==null ? k==null :
     * key.equals(k))}, then this method returns {@code v}; otherwise
     * it returns {@code null}.  (There can be at most one such mapping.)
     *
     * <p>A return value of {@code null} does not <i>necessarily</i>
     * indicate that the map contains no mapping for the key; it's also
     * possible that the map explicitly maps the key to {@code null}.
     * The {@link #containsKey containsKey} operation may be used to
     * distinguish these two cases.
     *
     * @see #put(Object, Object)
     */
	public Object get(Object key) {
		if (!checkKey(key))
			throw new IllegalArgumentException("Key is not a Lease object");
		return map.get(key);
	}

    /**
     * Associates the specified value with the specified key in this map.
     * If the map previously contained a mapping for the key, the old
     * value is replaced.
     *
     * @param key key with which the specified value is to be associated
     * @param value value to be associated with the specified key
     * @return the previous value associated with <tt>key</tt>, or
     *         <tt>null</tt> if there was no mapping for <tt>key</tt>.
     *         (A <tt>null</tt> return can also indicate that the map
     *         previously associated <tt>null</tt> with <tt>key</tt>.)
     */
	public Object put(Object key, Object value) {
		if (!checkKey(key))
			throw new IllegalArgumentException("Key is not a Lease object");
		if (!checkValue(value))
			throw new IllegalArgumentException("Value is not a Long object");
		return map.put((Lease)key, (Long)value);
	}

    /**
     * Removes the mapping for the specified key from this map if present.
     *
     * @param  key key whose mapping is to be removed from the map
     * @return the previous value associated with <tt>key</tt>, or
     *         <tt>null</tt> if there was no mapping for <tt>key</tt>.
     *         (A <tt>null</tt> return can also indicate that the map
     *         previously associated <tt>null</tt> with <tt>key</tt>.)
     */
	public Object remove(Object key) {
		if (!checkKey(key))
			throw new IllegalArgumentException("Key is not a Lease object");
		return map.remove(key);
	}

    /**
     * Copies all of the mappings from the specified map to this map.
     * These mappings will replace any mappings that this map had for
     * any of the keys currently in the specified map.
     *
     * @param m mappings to be stored in this map
     * @throws NullPointerException if the specified map is null
     */
	public void putAll(Map m) {
		map.putAll(m);
		
	}

    /**
     * Removes all of the mappings from this map.
     * The map will be empty after this call returns.
     */
	public void clear() {
		map.clear();		
	}

    /**
     * Returns a {@link Set} view of the keys contained in this map.
     * The set is backed by the map, so changes to the map are
     * reflected in the set, and vice-versa.  If the map is modified
     * while an iteration over the set is in progress (except through
     * the iterator's own <tt>remove</tt> operation), the results of
     * the iteration are undefined.  The set supports element removal,
     * which removes the corresponding mapping from the map, via the
     * <tt>Iterator.remove</tt>, <tt>Set.remove</tt>,
     * <tt>removeAll</tt>, <tt>retainAll</tt>, and <tt>clear</tt>
     * operations.  It does not support the <tt>add</tt> or <tt>addAll</tt>
     * operations.
     */
	public Set keySet() {
		return map.keySet();
	}

    /**
     * Returns a {@link Collection} view of the values contained in this map.
     * The collection is backed by the map, so changes to the map are
     * reflected in the collection, and vice-versa.  If the map is
     * modified while an iteration over the collection is in progress
     * (except through the iterator's own <tt>remove</tt> operation),
     * the results of the iteration are undefined.  The collection
     * supports element removal, which removes the corresponding
     * mapping from the map, via the <tt>Iterator.remove</tt>,
     * <tt>Collection.remove</tt>, <tt>removeAll</tt>,
     * <tt>retainAll</tt> and <tt>clear</tt> operations.  It does not
     * support the <tt>add</tt> or <tt>addAll</tt> operations.
     */
	public Collection values() {
		return map.values();
	}

    /**
     * Returns a {@link Set} view of the mappings contained in this map.
     * The set is backed by the map, so changes to the map are
     * reflected in the set, and vice-versa.  If the map is modified
     * while an iteration over the set is in progress (except through
     * the iterator's own <tt>remove</tt> operation, or through the
     * <tt>setValue</tt> operation on a map entry returned by the
     * iterator) the results of the iteration are undefined.  The set
     * supports element removal, which removes the corresponding
     * mapping from the map, via the <tt>Iterator.remove</tt>,
     * <tt>Set.remove</tt>, <tt>removeAll</tt>, <tt>retainAll</tt> and
     * <tt>clear</tt> operations.  It does not support the
     * <tt>add</tt> or <tt>addAll</tt> operations.
     *
     * @return a set view of the mappings contained in this map
     */
	public Set entrySet() {
		return map.entrySet();
	}

    /**
     * Returns true if the given object is a Lease which can be renewed
     * and cancelled in a batch with other leases in the map.  Whether
     * or not two Lease objects can be batched is an implementation
     * detail determined by the objects.
     *
     * @param key  an object that should be a lease
     * @return true if the lease can be renewed and cancelled in a 
     *              a batch with other leases
     *
     * @see Lease#canBatch
     */
	public boolean canContainKey(Object key) {
		if (checkKey(key))
			return true;
		return false;
	}

    /**
     * Renews all leases in the LeaseMap, using their current values as
     * the renewal durations.  If all renewals are successful, returns
     * normally.  Otherwise, removes all leases that failed to renew
     * from the LeaseMap, and throws LeaseMapException.
     * 
     * @throws LeaseMapException
     * @throws java.rmi.RemoteException
     */
	public void renewAll() throws LeaseMapException, RemoteException {
		Iterator<Lease> leases= map.keySet().iterator();
		Iterator<Long> values=map.values().iterator();
		Lease temp=null;
		Long duration=null;
		while (leases.hasNext()){
			try {				
				temp=leases.next();
				duration = values.next();
				temp.renew(duration);
			} catch (Exception e) {
				exceptionMap.put(temp, e);
			}
		}
		if (exceptionMap.size()!=0) {
			Iterator <Lease> it= exceptionMap.keySet().iterator();
			while (it.hasNext()){
				map.remove(it.next());
			}
			throw new LeaseMapException("There were some errors while renewing some leases", exceptionMap);
		}		
	}

    /**
     * Cancels all leases in the LeaseMap.  If all cancels are successful,
     * returns normally (leaving all leases in the map).  Otherwise,
     * removes all leases that failed to cancel from the LeaseMap, and
     * throws LeaseMapException.
     * 
     * @throws LeaseMapException
     * @throws java.rmi.RemoteException
     */
	public void cancelAll() throws LeaseMapException, RemoteException {
		Iterator<Lease> leases= map.keySet().iterator();
		Lease temp=null;
		while (leases.hasNext()){
			try {
				temp=leases.next();
				temp.cancel();
			} catch (Exception e) {
				exceptionMap.put(temp, e);
			}
		}
		if (exceptionMap.size()!=0) {
			Iterator <Lease> it= exceptionMap.keySet().iterator();
			while (it.hasNext()){
				map.remove(it.next());
			}
			throw new LeaseMapException("There were some errors while canceling some leases", exceptionMap);
		}
		
	}
	
	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	/** {@inheritDoc} */
	public String toString(){
		StringBuffer string = new StringBuffer();
		Iterator<Lease> leases= map.keySet().iterator();
		Iterator<Long> values=map.values().iterator();
		while(leases.hasNext()){
			string.append(leases.next()+" , "+values.next()+"\n");
		}
		if (exceptionMap.size()>0) {
			string.append("\nExceptions List\n");
			leases = exceptionMap.keySet().iterator();
			Iterator<Exception> e= exceptionMap.values().iterator();
			while (leases.hasNext()){
				string.append(leases.next()+ " , "+e.next().getMessage());
			}
		}
		return string.toString();
	}
	
	/**
	 * The main method.
	 *
	 * @param args the arguments
	 */
	public static void main(String [] args){
//		long duration = System.currentTimeMillis()+5000;
//		long dd=System.currentTimeMillis()+6000;
//		long renew=-9;
//		Lease lease = new Lease(duration);
//		Lease lease2 = new Lease(dd);
//		LeaseMap map= new LeaseMap(lease, renew);
//		map.put(lease2, Long.valueOf(200));
//
//		System.out.println(map);
//		try {
//		map.renewAll();
//	} catch (LeaseMapException e) {
////		System.out.println(e.exceptionMap);
//	} catch (RemoteException e) {
////		e.printStackTrace();
//	}
//		System.out.println(map);
//		
	}

}
