/*
 * 
 */
package casa.JION.space;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.PrintStream;
import java.rmi.MarshalledObject;
import java.rmi.RemoteException;
import java.security.SecureRandom;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import casa.JION.event.RemoteEvent;
import casa.JION.utile.DodwanSingleton;
import casa.dodwan.docware.Descriptor;
import casa.dodwan.docware.DescriptorZipBufferizer;
import casa.dodwan.docware.Payload;
import casa.dodwan.docware.PayloadBufferizer;
import casa.dodwan.message.Message;
import casa.dodwan.run.Dodwan;
import casa.dodwan.util.FileBasedRepository;
import casa.dodwan.util.Processor;
import casa.dodwan.util.StringGenerator;
import casa.dodwan.util.SystemEnvironment;

import net.jini.core.entry.Entry;
import net.jini.core.entry.UnusableEntryException;
import net.jini.core.event.EventRegistration;
import net.jini.core.event.RemoteEventListener;
import net.jini.core.lease.Lease;
import net.jini.core.transaction.Transaction;
import net.jini.core.transaction.TransactionException;

/**
 * This interface is implemented by servers that export a
 * JavaSpaces(TM) technology service.  The operations in this interface
 * are the public methods that all such spaces support.
 *
 * @author Sun Microsystems, Inc.
 *
 * @see Entry
 */
public class JavaSpace implements net.jini.space.JavaSpace{
	
	/** The dodwan. */
	private Dodwan dodwan=null;
	
	/** The lock. */
	protected static Object lock=null;
	
	/** The listeners. */
	protected static HashMap<Long,RemoteEventListener> listeners=null;
	
	/** The locks. */
	protected static HashMap<String, Object> locks=null;

	/** The space dir. */
	private FileBasedRepository<Descriptor> spaceDir=null;
	
	/** The space file. */
	private File spaceFile=null;
	
	/** The notify dir. */
	private FileBasedRepository<Descriptor> notifyDir=null;
	
	/** The notify file. */
	private File notifyFile=null;
	
	/** The handback dir. */
	private FileBasedRepository<Payload> handbackDir=null;
	
	/** The handback file. */
	private File handbackFile=null;
	
	/** The answer file. */
	private File answerFile=null;
	
	/** The answer dir. */
	private FileBasedRepository<Descriptor> answerDir=null;
	
	/** The lease file. */
	private File leaseFile=null;
	
	/** The lease dir. */
	private FileBasedRepository<Descriptor> leaseDir=null;
	
	/** The read file. */
	private File readFile =null;
	
	/** The read dir. */
	private FileBasedRepository<Descriptor> readDir=null;
	
	/** The entries. */
	private ArrayList<casa.JION.entry.Entry> entries = null;
	
	/** The event id. */
	private SecureRandom eventID=null;
	
	/**
	 * Instantiates a new java space.
	 */
	protected JavaSpace(){
		eventID=new SecureRandom();
		listeners= new HashMap<Long, RemoteEventListener>();
//		takes=new HashMap<Object, Descriptor>();
		locks=new HashMap<String, Object>();
		entries= new ArrayList<casa.JION.entry.Entry>();
		spaceFile = new File("space");
		spaceDir = new FileBasedRepository<Descriptor>(spaceFile, new DescriptorZipBufferizer());
		
		notifyFile = new File("notify");
		notifyDir = new FileBasedRepository<Descriptor>(notifyFile, new DescriptorZipBufferizer());
		
		handbackFile = new File("handback");
		handbackDir = new FileBasedRepository<Payload>(handbackFile, new PayloadBufferizer());
		
		answerFile = new File("answer");
		answerDir = new FileBasedRepository<Descriptor>(answerFile, new DescriptorZipBufferizer());
		
		leaseFile = new File("lease");
		leaseDir = new FileBasedRepository<Descriptor>(leaseFile, new DescriptorZipBufferizer());
		
		readFile = new File("read");
		readDir = new FileBasedRepository<Descriptor>(readFile, new DescriptorZipBufferizer());
		
		outOffDate();
		File[] files = spaceFile.listFiles();
		Descriptor temp=null;
		for (File x:files){
			temp=spaceDir.get(x.getName());
			entries.add(new casa.JION.entry.Entry(temp));
		}
//		System.out.println(entries.size());
//		for(int i=0;i<entries.size();i++){
//			System.out.println(entries.get(i));
//		}
		dodwan= DodwanSingleton.getDodwan();
		if (! dodwan.pubSubService.getKeys().contains("JavaSpace")){
			Descriptor pattern=new Descriptor();
			pattern.setAttribute("SERVICE", "TUPLESPACE");
			dodwan.pubSubService.addSubscription("JavaSpace", pattern, new MyProcessor());
		}
		
		if (! dodwan.pubSubService.getKeys().contains("Notify")){
			Descriptor pattern=new Descriptor();
			pattern.setAttribute("NOTIFY", "OPERATION");
			dodwan.pubSubService.addSubscription("Notify", pattern, new MyProcessor_Notify());
		}
	}

//	private Message prepareForDodwan(casa.tuplespaces.entry.Entry entry, long duration){
//		Message msg= new Message(entry.getDescriptor(),entry.getPayload());
//		msg.getDescriptor().setDeadline(new Date(duration));
//		msg.getDescriptor().setAttribute("cachable", "true");
//		msg.getDescriptor().setUniqueDocumentId();
//		msg.getDescriptor().setSource(SystemEnvironment.host);
//		return msg;
//	}
	
	/**
 * Adds the entry.
 *
 * @param entry the entry
 * @param lease the lease
 */
private void addEntry(casa.JION.entry.Entry entry, casa.JION.lease.Lease lease){
		this.entries.add(entry);
		spaceDir.put(entry.getKey(), entry.getDescriptor());
		leaseDir.put(lease.getEntryID(),lease.leaseToDescriptor());
		System.out.println(entry.getKey()+" has been added to the space");
		outOffDate();
	}
	
	/**
	 * this function will search all the local read requests in order to find those which match the entry.
	 * Then it will send this entry to the sources of these requests "Hosts"
	 *
	 * @param entry the entry
	 */
	private void checkRead(casa.JION.entry.Entry entry){
		outOffDate();
		if (entry ==null)
			throw new IllegalArgumentException("Entry can not be null");
		Descriptor desc=entry.getDescriptor();
		Descriptor pattern=null;
		String key=null;
		Date deadline=null;
		for (File x: readFile.listFiles()){
			key = x.getName();
			pattern=readDir.get(key);
			deadline=pattern.getDeadline();
			pattern.removeAttribute("_dline");
			if (desc.matches(pattern)){
				readDir.delete(key);
				Message message=new Message(entry);
				message.getDescriptor().setAttribute("cachable", "true");
				message.getDescriptor().setUniqueDocumentId();
				message.getDescriptor().setSource(SystemEnvironment.host);
				message.getDescriptor().setAttribute("SERVICE", "TUPLESPACE");
				message.getDescriptor().setAttribute("TYPE", "ANSWER");
				message.getDescriptor().setAttribute("RID", key);
				if (deadline!=null)
					message.getDescriptor().setDeadline(deadline);
				message.setPayload("TupleSpaceRead".getBytes());
				System.out.println("An answer has been sended to "+key);
				System.out.println(message.getDescriptor());
				try {
					dodwan.pubSubService.publish(message);
				} catch (Exception e) {
					e.printStackTrace();
				}
			}
			
		}
		
	}
	
    /**
     * Write a new entry into the space.
     *
     * @param entry  the entry to write
     * @param txn the transaction object, if any, under which to 
     *            perform the write
     * @param lease  the requested lease time, in milliseconds
     * @return a lease for the entry that was written to the space 
     *
     * @throws TransactionException if a transaction error occurs
     * @throws RemoteException if a communication error occurs
     */
	public Lease write(Entry entry, Transaction txn, long lease)
			throws TransactionException, RemoteException {
		casa.JION.entry.Entry myEntry= (casa.JION.entry.Entry) entry;
//		if (checkDesc(myEntry.getDescriptor())){
//			System.out.println("The descriptor has been consumed by a take function");
//			return null;
//		}
		checkRead(myEntry);
		myEntry.getDescriptor().setUniqueDocumentId();
		long result=System.currentTimeMillis()+lease;
		if ((lease>=0)&&(result<=0)){
			result=lease;
		}
		casa.JION.lease.Lease leaseReturned=new casa.JION.lease.Lease(myEntry.getKey(),result);
		myEntry.getDescriptor().setDeadline(new Date(leaseReturned.getExpiration()));
//		myEntry.setLease(leaseReturned);
		myEntry.getDescriptor().setSource(SystemEnvironment.host);
		myEntry.getDescriptor().setAttribute("cachable", "true");
		addEntry(myEntry,leaseReturned);
		myEntry.getDescriptor().setAttribute("NOTIFY", "OPERATION");
		Message msg = new Message(myEntry);
		try {
			System.out.println(msg.getDescriptor());
			System.out.println("The entry has been published by DoDWAN //notify purposes");
			dodwan.pubSubService.publish(msg);
		} catch (Exception e) {
			e.printStackTrace();
		}
//		if (Configuration.immutatedTuple) {
//			try {
//				myEntry.getDescriptor().setAttribute("tuple", "space");
//				Message msg = new Message(myEntry);
//				System.out.println(msg.getDescriptor());
//				System.out.println(new String(msg.getPayload()).trim());
//				dodwan.pubSubService.publish(msg);
//				System.out.println("The entry has been published by DoDWAN");
//			} catch (Exception e) {
//				e.printStackTrace();
//			}
//		}
		return leaseReturned;
	}
	
	/**
	 * this function could be called only by write function. It will verify if a take operation is already waiting for such type of descriptor
	 *
	 * @param tmpl the tmpl
	 * @param txn the txn
	 * @param timeout the timeout
	 * @return true if the descriptor is consumed by a waiting take. False otherwise
	 * @throws UnusableEntryException the unusable entry exception
	 * @throws TransactionException the transaction exception
	 * @throws InterruptedException the interrupted exception
	 * @throws RemoteException the remote exception
	 */
//	private boolean checkDesc(Descriptor desc){
//		boolean checked=false;
//		Descriptor temp;
//		Object lock;
//		Iterator locks= takes.keySet().iterator();
//		for (int i=0; i<takes.size();i++){
//			lock = locks.next();
//			temp=takes.get(lock);
//			if (desc.matches(temp)){
//				takes.remove(lock);
//				answerDir.put(lock.toString(), desc);
////				synchronized(lock){
////					lock.notify();
////				}
//				latch.countDown();
//				
//			}
//		}
//		return checked;
//	}
	  /**
     * Take a matching entry from the space, waiting until one exists.
     * Matching is and timeout done as for <code>read</code>.
     *
     * @param tmpl      The template used for matching.  Matching is
     *			done against <code>tmpl</code> with <code>null</code>
     *			fields being wildcards ("match anything") other
     *			fields being values ("match exactly on the
     *			serialized form").
     * @param txn	The transaction (if any) under which to work.
     * @param timeout	How long the client is willing to wait for a
     *			transactionally proper matching entry.  A
     *			timeout of <code>NO_WAIT</code> means to wait
     *			no time at all; this is equivalent to a wait
     *			of zero.
     *
     * @return the entry taken from the space
     * @throws UnusableEntryException if any serialized field of the entry 
     *         being read cannot be deserialized for any reason 
     * @throws TransactionException if a transaction error occurs
     * @throws InterruptedException if the thread in which the take
     *         occurs is interrupted
     * @throws RemoteException if a communication error occurs
     * @throws IllegalArgumentException if a negative timeout value is used
     * @see #read
     */
	public Entry take(Entry tmpl, Transaction txn, long timeout)
			throws UnusableEntryException, TransactionException,
			InterruptedException, RemoteException {
		casa.JION.entry.Entry entryParm=(casa.JION.entry.Entry) tmpl;
		casa.JION.entry.Entry entryAns=null;
		outOffDate();
		boolean finish=false;
		Descriptor temp=null;
		for (File x:spaceFile.listFiles()){
			temp=spaceDir.get(x.getName());
			if(temp.matches(entryParm.getDescriptor())){
				entryAns=new casa.JION.entry.Entry();
				entryAns.setDescriptor(temp);
				finish=true;
				spaceDir.delete(x.getName());
				leaseDir.delete(x.getName());
				for (int i=0; i< entries.size();i++){
					if (entries.get(i).getKey().equalsIgnoreCase(x.getName())){
						entries.remove(i);
					}
				}
				break;
			}
		}
		if (!finish&& timeout>NO_WAIT){
			Date limit= new Date(System.currentTimeMillis()+timeout);
			casa.JION.entry.Entry entry = (casa.JION.entry.Entry)read(tmpl, txn, timeout);
			if (entry !=null){
				long timeout2= limit.getTime()-System.currentTimeMillis();
				Descriptor desc= new Descriptor();
				desc.setAttribute("cachable", "true");
				desc.setSource(SystemEnvironment.host);
				desc.setUniqueDocumentId();
				desc.setDeadline(limit);
				String RID= new StringGenerator().newString();
				desc.setAttribute("RID", RID);
				desc.setAttribute("SERVICE", "TUPLESPACE");
				desc.setAttribute("TYPE","TAKE");
				desc.setAttribute("PHASE", "ASK");
				desc.setAttribute("FROM",entry.getDescriptor().getSource());
				StringBuffer selector=new StringBuffer();
				Iterator<Map.Entry<String, String>> it=entryParm.getDescriptor().getEntrySet().iterator();
				if (it.hasNext()){
					selector.append(it.next());
				}
				while(it.hasNext()){
					selector.append(","+it.next());
				}
				desc.setAttribute("SELECTOR", selector.toString());
//				desc.setAttribute("WHICH", entry.getKey());
				lock=new Object();
				locks.put(RID, lock);
				Message msg=new Message(desc,"JavaSpace".getBytes());
				try {
					System.out.println(msg.getDescriptor());
					System.out.println(new String(msg.getPayload()).trim());
					dodwan.pubSubService.publish(msg);
					System.out.println("DoDWAN will search the ok for the entry");
				} catch (Exception e) {
					e.printStackTrace();
				}
//				File answer=new File(answerFile,lock.toString());
				synchronized (lock) {
					System.out.println("Waiting");
					lock.wait(timeout2);
				}

				entryAns=getAnswer(RID);
			}
		}
		return entryAns;
	}
	
	
	
    /**
     * Read any matching entry from the space, blocking until one exists.
     * Return <code>null</code> if the timeout expires.
     *
     * @param tmpl      The template used for matching.  Matching is
     *			done against <code>tmpl</code> with <code>null</code>
     *			fields being wildcards ("match anything") other
     *			fields being values ("match exactly on the
     *			serialized form").
     * @param txn	The transaction (if any) under which to work.
     * @param timeout	How long the client is willing to wait for a
     *			transactionally proper matching entry.  A
     *			timeout of <code>NO_WAIT</code> means to wait
     *			no time at all; this is equivalent to a wait
     *			of zero.
     *
     * @return a copy of the entry read from the space
     * @throws UnusableEntryException if any serialized field of the entry 
     *         being read cannot be deserialized for any reason 
     * @throws TransactionException if a transaction error occurs
     * @throws InterruptedException if the thread in which the read
     *         occurs is interrupted
     * @throws RemoteException if a communication error occurs
     * @throws IllegalArgumentException if a negative timeout value is used
     */
	public Entry read(Entry tmpl, Transaction txn, long timeout)
			throws UnusableEntryException, TransactionException,
			InterruptedException, RemoteException {
		outOffDate();
		boolean finish=false;
		casa.JION.entry.Entry entryParm=(casa.JION.entry.Entry) tmpl;
		casa.JION.entry.Entry entryAns=null;
		Descriptor temp=null;
//		System.out.println(entryParm.getDescriptor());
		for(File x : spaceFile.listFiles()){
			temp=spaceDir.get(x.getName());
			if (temp.matches(entryParm.getDescriptor())){
				entryAns=new casa.JION.entry.Entry();
				entryAns.setDescriptor(temp);
				finish=true;
				break;
			}
		}
		if (!finish&& timeout>NO_WAIT){
			String RID= new StringGenerator().newString();
			StringBuffer selector=new StringBuffer();
			Iterator<Map.Entry<String, String>> it=entryParm.getDescriptor().getEntrySet().iterator();
			if (it.hasNext()){
				selector.append(it.next());
			}
			while(it.hasNext()){
				selector.append(","+it.next());
			}
			Descriptor desc= new Descriptor();
			desc.setAttribute("cachable", "true");
			desc.setSource(SystemEnvironment.host);
			desc.setUniqueDocumentId();
			desc.setAttribute("SELECTOR", selector.toString());
			desc.setDeadline(new Date(System.currentTimeMillis()+timeout));
			desc.setAttribute("SERVICE", "TUPLESPACE");
			desc.setAttribute("TYPE","READ");
			desc.setAttribute("RID",RID);
			lock=new Object();
			locks.put(RID, lock);
			Message msg=new Message(desc,"JavaSpace".getBytes());
			try {
				System.out.println(msg.getDescriptor());
				System.out.println(new String(msg.getPayload()).trim());
				dodwan.pubSubService.publish(msg);
				System.out.println("DoDWAN will search for that entry");
			} catch (Exception e) {
				e.printStackTrace();
			}
//			File answer=new File(answerFile,RID);
			synchronized (lock) {
				System.out.println("Waiting");
				System.out.println(timeout);
				lock.wait(timeout);		
			}
			entryAns=getAnswer(RID);
		}
		return entryAns;
	}
	
	/**
	 * Gets the answer.
	 *
	 * @param RID the rID
	 * @return the answer
	 */
	private casa.JION.entry.Entry getAnswer(String RID){
		casa.JION.entry.Entry entry=null;
		Descriptor desc= answerDir.get(RID);
		if (desc!=null){
			entry=new casa.JION.entry.Entry();
			entry.setDescriptor(desc);
			answerDir.delete(RID);
		}
		return entry;
	}
	

	  /**
     * Read any matching entry from the space, returning
     * <code>null</code> if there is currently is none.  Matching and
     * timeouts are done as in <code>read</code>, except that blocking
     * in this call is done only if necessary to wait for transactional
     * state to settle.
     *
     * @param tmpl      The template used for matching.  Matching is
     *			done against <code>tmpl</code> with <code>null</code>
     *			fields being wildcards ("match anything") other
     *			fields being values ("match exactly on the
     *			serialized form").
     * @param txn	The transaction (if any) under which to work.
     * @param timeout	How long the client is willing to wait for a
     *			transactionally proper matching entry.  A
     *			timeout of <code>NO_WAIT</code> means to wait
     *			no time at all; this is equivalent to a wait
     *			of zero.
     *
     * @return a copy of the entry read from the space
     * @throws UnusableEntryException if any serialized field of the entry 
     *         being read cannot be deserialized for any reason 
     * @throws TransactionException if a transaction error occurs
     * @throws InterruptedException if the thread in which the read
     *         occurs is interrupted
     * @throws RemoteException if a communication error occurs
     * @throws IllegalArgumentException if a negative timeout value is used
     * @see #read
     */
	public Entry readIfExists(Entry tmpl, Transaction txn, long timeout)
			throws UnusableEntryException, TransactionException,
			InterruptedException, RemoteException {
		return read(tmpl, txn, NO_WAIT);
	}



    /**
     * Take a matching entry from the space, returning
     * <code>null</code> if there is currently is none.  Matching is
     * and timeout done as for <code>read</code>, except that blocking
     * in this call is done only if necessary to wait for transactional
     * state to settle.
     *
     * @param tmpl      The template used for matching.  Matching is
     *			done against <code>tmpl</code> with <code>null</code>
     *			fields being wildcards ("match anything") other
     *			fields being values ("match exactly on the
     *			serialized form").
     * @param txn	The transaction (if any) under which to work.
     * @param timeout	How long the client is willing to wait for a
     *			transactionally proper matching entry.  A
     *			timeout of <code>NO_WAIT</code> means to wait
     *			no time at all; this is equivalent to a wait
     *			of zero.
     *
     * @return the entry taken from the space
     * @throws UnusableEntryException if any serialized field of the entry 
     *         being read cannot be deserialized for any reason 
     * @throws TransactionException if a transaction error occurs
     * @throws InterruptedException if the thread in which the take
     *         occurs is interrupted
     * @throws RemoteException if a communication error occurs
     * @throws IllegalArgumentException if a negative timeout value is used
     * @see #read
     */
	public Entry takeIfExists(Entry tmpl, Transaction txn, long timeout)
			throws UnusableEntryException, TransactionException,
			InterruptedException, RemoteException {
		return take(tmpl,txn,NO_WAIT);
	}


    /**
     * When entries are written that match this template notify the
     * given <code>listener</code> with a <code>RemoteEvent</code> that
     * includes the <code>handback</code> object.  Matching is done as
     * for <code>read</code>.
     *
     * @param tmpl      The template used for matching.  Matching is
     *			done against <code>tmpl</code> with <code>null</code>
     *			fields being wildcards ("match anything") other
     *			fields being values ("match exactly on the
     *			serialized form").
     * @param txn	The transaction (if any) under which to work.
     * @param listener  The remote event listener to notify.
     * @param lease  the requested lease time, in milliseconds
     * @param handback  An object to send to the listener as part of the 
     *                  event notification.
     * @return the event registration to the the registrant
     * @throws TransactionException if a transaction error occurs
     * @throws RemoteException if a communication error occurs
     * @throws IllegalArgumentException if the lease time requested 
     *         is not Lease.ANY and is negative
     * @see #read
     * @see net.jini.core.event.EventRegistration
     */
	public EventRegistration notify(Entry tmpl, Transaction txn,
			RemoteEventListener listener, long lease, MarshalledObject handback)
			throws TransactionException, RemoteException {
		long id=eventID.nextLong();
		String source=SystemEnvironment.host;
		long seq=0;
		long result=System.currentTimeMillis()+lease;
//		System.out.println(lease+" "+result);
		if ((lease>=0)&&(result<=0)){
			result=lease;
		}
		casa.JION.lease.Lease myLease=new casa.JION.lease.Lease(String.valueOf(id), result);
		casa.JION.entry.Entry myEntry = (casa.JION.entry.Entry) tmpl;
		Descriptor desc = new Descriptor(myEntry.getDescriptor());
		desc.setDeadline(new Date(myLease.getExpiration()));
		desc.setAttribute("SEQ", String.valueOf(seq));
		desc.setAttribute("EVENTID", String.valueOf(id));
		desc.setAttribute("SOURCE", source);
		try {
			ByteArrayOutputStream out=new ByteArrayOutputStream();
			ObjectOutputStream objectOutputStream=new ObjectOutputStream(out);
			objectOutputStream.writeObject(handback);
			out.flush();
			objectOutputStream.flush();
			byte [] myHandback=out.toByteArray();
			handbackDir.put(String.valueOf(id), new Payload(myHandback));
		} catch (IOException e) {
			e.printStackTrace();
		}
		notifyDir.put(String.valueOf(id), desc);
		leaseDir.put(String.valueOf(id), myLease.leaseToDescriptor());
		listeners.put(id,listener);
		
		casa.JION.event.EventRegistration eventRegistration=new casa.JION.event.EventRegistration(id, source, myLease, seq);
		return eventRegistration;
	}

    /**
     * The process of serializing an entry for transmission to a JavaSpaces
     * service will be identical if the same entry is used twice. This is most
     * likely to be an issue with templates that are used repeatedly to search
     * for entries with <code>read</code> or <code>take</code>. The client-side 
     * implementations of <code>read</code> and <code>take</code> cannot
     * reasonably avoid this duplicated effort, since they have no efficient
     * way of checking whether the same template is being used without
     * intervening modification.
     *
     * The <code>snapshot</code> method gives the JavaSpaces service implementor
     * a way to reduce the impact of repeated use of the same entry. Invoking
     * <code>snapshot</code> with an <code>Entry</code> will return another
     * <code>Entry</code> object that contains a <i>snapshot</i> of the
     * original entry. Using the returned snapshot entry is equivalent to
     * using the unmodified original entry in all operations on the same
     * JavaSpaces service. Modifications to the original entry will not
     * affect the snapshot. You can <code>snapshot</code> a <code>null</code>
     * template; <code>snapshot</code> may or may not return null given a
     * <code>null</code> template.
     *
     * The entry returned from <code>snapshot</code> will be guaranteed
     * equivalent to the original unmodified object only when used with
     * the space. Using the snapshot with any other JavaSpaces service
     * will generate an <code>IllegalArgumentException</code> unless the
     * other space can use it because of knowledge about the JavaSpaces
     * service that generated the snapshot. The snapshot will be a different
     * object from the original, may or may not have the same hash code,
     * and <code>equals</code> may or may not return <code>true</code>
     * when invoked with the original object, even if the original
     * object is unmodified.
     *
     * A snapshot is guaranteed to work only within the virtual machine
     * in which it was generated. If a snapshot is passed to another 
     * virtual machine (for example, in a parameter of an RMI call),
     * using it--even with the same JavaSpaces service--may generate
     * an <code>IllegalArgumentException</code>.
     *
     * @param e the entry to take a snapshot of.
     * @return a snapshot of the entry.
     * @throws RemoteException if a communication error occurs
     */
	public Entry snapshot(Entry e) throws RemoteException {
		outOffDate();
		casa.JION.entry.Entry myEntry= (casa.JION.entry.Entry)e;
		return new casa.JION.entry.Entry(myEntry.getDescriptor());
	}
	
	/**
	 * This function will print a list of entries exist on the local space.
	 *
	 * @param out the out
	 */
	public void listEntries(PrintStream out){
		outOffDate();
		for (int i=0; i< entries.size();i++){
			out.println(entries.get(i).getKey()+" , "+entries.get(i).getDescriptor().getDeadline());
		}
	}
	
	/**
	 * This function will print a list of entries exist on the local space.
	 *
	 * @param out the out
	 */
	public void listNotifies(PrintStream out){
		outOffDate();
		Descriptor desc=null;
		for (File x : notifyFile.listFiles()){
			desc = notifyDir.get(x.getName());
			out.println(x.getName()+" , "+desc.getDeadline());
		}
	}
	
	/**
	 * This function will print all the fields of a specific entry.
	 *
	 * @param id Entry's id
	 * @return the casa. jio n.entry. entry
	 */
	public casa.JION.entry.Entry detail(String id){
		outOffDate();
		casa.JION.entry.Entry entry=null;
		Descriptor desc = spaceDir.get(id);
		if (desc!=null) {
			entry=new casa.JION.entry.Entry();
			entry.setDescriptor(desc);
		}
		return entry;
	}
	
	/**
	 * This function can be called only by Lease.cancel() function. It will cancel the specific lease
	 *
	 * @param lease the lease
	 */
	public void doCancel(casa.JION.lease.Lease lease){
		outOffDate();
		boolean notify=false;
		Descriptor desc =null;
		if (spaceDir.contains(lease.getEntryID()))
			desc = spaceDir.get(lease.getEntryID());
		if (desc==null){
			desc=notifyDir.get(lease.getEntryID());
			if (desc!=null)
				notify=true;
		}
		if (desc==null)
			throw new IllegalArgumentException("There is no such entry in the space");
		leaseDir.delete(lease.getEntryID());
		if (!notify){
			spaceDir.delete(lease.getEntryID());
			for (int i=0; i< entries.size();i++){
				if (entries.get(i).getKey().equalsIgnoreCase(lease.getEntryID())){
					entries.remove(i);
				}
			}
		}
		else {
			notifyDir.delete(lease.getEntryID());
			listeners.remove(lease.getEntryID());
			handbackDir.delete(lease.getEntryID());
		}
	}
	
	/**
	 * This function can be called only by Lease.renew() function. It will renew the expiration for the entry specified by @param id
	 *
	 * @param lease the lease
	 */
	public void doRenew(casa.JION.lease.Lease lease){
		outOffDate();
		boolean notify=false;
		Descriptor desc =null;
		if (spaceDir.contains(lease.getEntryID()))
			desc = spaceDir.get(lease.getEntryID());
		if (desc==null){
			desc=notifyDir.get(lease.getEntryID());
			System.out.println("desc: "+desc);
			if (desc!=null)
				notify=true;
		}
		if (desc==null)
			throw new IllegalArgumentException("There is no such entry in the space");
//		System.out.println("desc before: "+desc);
//		System.out.println("Lease before: "+new casa.tuplespaces.lease.Lease(leaseDir.get(lease.getEntryID())));
		leaseDir.delete(lease.getEntryID());
		leaseDir.put(lease.getEntryID(), lease.leaseToDescriptor());
		desc.setDeadline(new Date(lease.getExpiration()));
		if (!notify) {
			spaceDir.delete(lease.getEntryID());
			spaceDir.put(lease.getEntryID(), desc);
		}
		else
		{
			notifyDir.delete(lease.getEntryID());
			notifyDir.put(lease.getEntryID(), desc);
		}
		if (!notify) {
			//		System.out.println("desc after: "+desc);
			//		System.out.println("Lease after: "+lease);
			for (int i = 0; i < entries.size(); i++) {
				if (entries.get(i).getKey()
						.equalsIgnoreCase(lease.getEntryID())) {
					//				System.out.println("Entry before: "+entries.get(i));
					//				entries.get(i).getDescriptor().setDeadline(new Date(lease.getExpiration()));
					casa.JION.entry.Entry element = new casa.JION.entry.Entry();
					element.setDescriptor(desc);
					entries.set(i, element);
					//				System.out.println("Entry after: "+entries.get(i));
				}
			}
			//		Iterator<casa.tuplespaces.entry.Entry> it= entries.iterator();
			//		casa.tuplespaces.entry.Entry entry=null;
			//		while(it.hasNext()){
			//			entry=it.next();
			//			if (entry.getKey().equalsIgnoreCase(lease.getEntryID())){
			//				System.out.println("Enties before: "+entry);
			//				it.remove();
			//				entry.setDescriptor(desc);
			//				break;
			//			}
			//		}
			//		casa.tuplespaces.entry.Entry e=new casa.tuplespaces.entry.Entry();
			//		e.setDescriptor(desc);
			//		System.out.println("Enties after: "+e);
			//		entries.add(e);
		}
	}
	
	/**
	 * This function retrieve the specific lease from the lease directory.
	 *
	 * @param id the id
	 * @return the lease
	 */
	public casa.JION.lease.Lease getLease(String id){
		outOffDate();
		casa.JION.lease.Lease lease=null;
		Descriptor desc=leaseDir.get(id);
		if (desc!=null){
			lease=new casa.JION.lease.Lease(desc);
		}
		return lease;
	}
	
	/**
	 * Out off date.
	 */
	public void outOffDate(){
		Date now = new Date(System.currentTimeMillis());
		Date temp=null;
		Iterator<casa.JION.entry.Entry> it= entries.iterator();
		casa.JION.entry.Entry entry=null;
		while (it.hasNext()){
			entry=it.next();
			temp=entry.getDescriptor().getDeadline();
			if (temp!=null&&now.compareTo(temp)>0){
				it.remove();
				spaceDir.delete(entry.getKey());
				leaseDir.delete(entry.getKey());
				System.out.println(entry.getKey()+" has been deleted");
			}
		}
		
		Descriptor desc=null;
		for(File x:notifyFile.listFiles()){
			desc=notifyDir.get(x.getName());
			temp=desc.getDeadline();
			if (temp!=null&&now.compareTo(temp)>0){
				leaseDir.delete(x.getName());
				notifyDir.delete(x.getName());
				handbackDir.delete(x.getName());
				listeners.remove(x.getName());
				System.out.println(x.getName()+" has been deleted");
			}
		}
		
		desc=null;
		for(File x:readFile.listFiles()){
			desc=readDir.get(x.getName());
			temp=desc.getDeadline();
			if (temp!=null&&now.compareTo(temp)>0){
				readDir.delete(x.getName());
				System.out.println(x.getName()+" has been deleted");
			}
		}
	}
	
	/**
	 * String to descriptor.
	 *
	 * @param str the str
	 * @return the descriptor
	 * @throws IllegalArgumentException the illegal argument exception
	 */
	private Descriptor stringToDescriptor(String str) throws IllegalArgumentException{
		Descriptor desc=new Descriptor();
		String[] patterns=str.split(",");
		for (int i =0; i< patterns.length;i++) {
			String[] tpattern=patterns[i].split("=");
			if(tpattern.length!=2)
				throw new IllegalArgumentException("Selector format is wrong. It must be attribut=value,attribut=value...");
			desc.setAttribute(tpattern[0], (tpattern[1].equalsIgnoreCase("null"))?".*":tpattern[1]);
		}
		return desc;
	}
	
	/**
	 * this function will update the seq field for the given pattern and save it.
	 *
	 * @param key the key
	 * @param pattern_original the pattern_original
	 * @return the remote event
	 */
	private RemoteEvent newEvent(String key, Descriptor pattern_original){
		notifyDir.delete(key);
		Descriptor result=new Descriptor(pattern_original);
		long seqNum=Long.valueOf(pattern_original.getAttribute("SEQ")).longValue();
		long eventID=Long.valueOf(pattern_original.getAttribute("EVENTID")).longValue();
		String source=pattern_original.getAttribute("SOURCE");
		MarshalledObject<Object> handback=null;
		try {
				Payload payload = handbackDir.get(key);
				ByteArrayInputStream in = new ByteArrayInputStream(payload.getBytes());
				ObjectInputStream ois= new ObjectInputStream(in);
				handback=(MarshalledObject) ois.readObject();
//				System.out.println(object.hashCode());
//				System.out.println(handback.hashCode());
		} catch (Exception e) {
			e.printStackTrace();
		}
		
		seqNum=seqNum+1;
		result.removeAttribute("SEQ");
		result.setAttribute("SEQ", String.valueOf(seqNum));
		notifyDir.put(key, result);
		return new RemoteEvent(source, eventID, seqNum, handback);
	}
	
	/**
	 * The Class MyProcessor_Notify.
	 */
	public class MyProcessor_Notify implements Processor<Message> {

		/* (non-Javadoc)
		 * @see casa.dodwan.util.Processor#process(java.lang.Object)
		 */
		/** {@inheritDoc} */
		public void process(Message msg) throws Exception {
			outOffDate();
			Descriptor desc = msg.getDescriptor();
			Descriptor pattern_original=null;
			Descriptor pattern_template=null;
			String key = null;
			for (File x : notifyFile.listFiles()){
				key = x.getName();
				pattern_original= notifyDir.get(key);
				pattern_template=new Descriptor(pattern_original);
				pattern_template.removeAttribute("SEQ");
				pattern_template.removeAttribute("EVENTID");
				pattern_template.removeAttribute("SOURCE");
				pattern_template.removeAttribute("_dline");
				if (desc.matches(pattern_template)){
					RemoteEvent event = newEvent(key, pattern_original);
					RemoteEventListener remoteEventListener=null;
					remoteEventListener=listeners.get(key);
					if (remoteEventListener == null){
						remoteEventListener = new casa.JION.event.RemoteEventListener();
					}
					remoteEventListener.notify(event);
				}
			}
			
		}
		
	}
	
	/**
	 * This function will delete the local read request that has been answered be another host.
	 *
	 * @param key the key
	 */
	private void chechAnswer (String key){
		outOffDate();
		for (File x:readFile.listFiles()){
			if (x.getName().equalsIgnoreCase(key))
				readDir.delete(key);
		}
	}
	
	/**
	 * The Class MyProcessor.
	 */
	public class MyProcessor implements Processor<Message> {

		/* (non-Javadoc)
		 * @see casa.dodwan.util.Processor#process(java.lang.Object)
		 */
		/** {@inheritDoc} */
		public void process(Message msg) throws Exception {
			System.out.println("New message has been arrived");
			Descriptor desc= msg.getDescriptor();
			String RID= desc.getAttribute("RID");
			String type=desc.getAttribute("TYPE");
			Date deadline= desc.getDeadline();
			desc.removeAttribute("RID");
			desc.removeAttribute("SERVICE");
			desc.removeAttribute("TYPE");
			if (RID!=null&&type!=null){
				if (type.equalsIgnoreCase("ANSWER")){
					System.out.println("ANSWER");
					chechAnswer(RID);
					if (locks.containsKey(RID)){
						Object lock=locks.get(RID);
						answerDir.put(RID, desc);
						synchronized(lock){
							lock.notify();
						}
					}
					else
						System.out.println("This answer is not for me");
				}
				if (type.equalsIgnoreCase("READ")){
					System.out.println("READ");
					casa.JION.entry.Entry tmpl=new casa.JION.entry.Entry();
					tmpl.setDescriptor(stringToDescriptor(desc.getAttribute("SELECTOR")));
					System.out.println(tmpl);
					casa.JION.entry.Entry entry= (casa.JION.entry.Entry)read(tmpl,null,NO_WAIT);
					if (entry!=null){
						Message message=new Message(entry);
						message.getDescriptor().setAttribute("cachable", "true");
						message.getDescriptor().setUniqueDocumentId();
						message.getDescriptor().setSource(SystemEnvironment.host);
						message.getDescriptor().setAttribute("SERVICE", "TUPLESPACE");
						message.getDescriptor().setAttribute("TYPE", "ANSWER");
						message.getDescriptor().setAttribute("RID", RID);
						if (deadline!=null)
							message.getDescriptor().setDeadline(deadline);
						message.setPayload("TupleSpaceRead".getBytes());
						System.out.println("An answer has been sended to "+RID);
						System.out.println(message.getDescriptor());
						dodwan.pubSubService.publish(message);
						
					} else {
						tmpl.getDescriptor().setDeadline(deadline);
						readDir.put(RID, tmpl.getDescriptor());
					}
				}
				if (type.equalsIgnoreCase("TAKE")){
					System.out.println("TAKE");
					String phase = desc.getAttribute("PHASE");
					if (phase.equalsIgnoreCase("ASK")){
						System.out.println("ASK");
						String from= desc.getAttribute("FROM");
						if (from.equalsIgnoreCase(SystemEnvironment.host)){
							System.out.println("FROM ME");
							casa.JION.entry.Entry tmpl=new casa.JION.entry.Entry();
							tmpl.setDescriptor(stringToDescriptor(desc.getAttribute("SELECTOR")));
							casa.JION.entry.Entry entry= (casa.JION.entry.Entry)take(tmpl,null,NO_WAIT);
							if (entry!=null){
								Message message=new Message(entry);
								message.getDescriptor().setAttribute("SERVICE", "TUPLESPACE");
								message.getDescriptor().setAttribute("TYPE", "ANSWER");
								message.getDescriptor().setAttribute("RID", RID);
								message.getDescriptor().setAttribute("cachable", "true");
								message.getDescriptor().setUniqueDocumentId();
								message.getDescriptor().setSource(SystemEnvironment.host);
								if (deadline!=null)
									message.getDescriptor().setDeadline(deadline);
								message.setPayload("TupleSpaceTake".getBytes());
								System.out.println("An answer has been sended to "+RID);
								System.out.println(message.getDescriptor());
								dodwan.pubSubService.publish(message);
								
							}
						}
						
					}

				}
				
			}
			
		}
		
	}
	
	/**
	 * The main method.
	 *
	 * @param args the arguments
	 */
	public static void main(String [] args){
		JavaSpace js=new JavaSpace();
//		casa.tuplespaces.entry.Entry entry = new casa.tuplespaces.entry.Entry();
//		entry.addField("gogo", "null");
		casa.JION.entry.Entry tmpl = new casa.JION.entry.Entry();
//		tmpl.addField("gogo", "kader200");
		tmpl.addField("ss", "dd");
		try {
//			MarshalledObject handback=new MarshalledObject(null);
//			casa.tuplespaces.event.EventRegistration er= (casa.tuplespaces.event.EventRegistration) js.notify(tmpl, null, null, 2*24*3600000, handback);
//			System.out.println(er);
//			System.out.println(js.listeners.size());
//			Lease lease=js.write(tmpl, null, 1*24*3600000);
//			Thread.sleep(2566);
//			casa.tuplespaces.lease.Lease lease=js.getLease("-1022984382057486939");
			casa.JION.entry.Entry entry = (casa.JION.entry.Entry)js.take(tmpl, null, 50000000);
//			Lease lease=js.write(tmpl, null, 1*24*3600000);
//			System.out.println(lease);
			System.out.println(entry);
//			lease.cancel();
//			lease.renew(1*24*3600000);
//			lease.renew(111);
//			lease=js.getLease("7804137902473502358");
//			System.out.println(lease);
//			lease.cancel();
//			System.out.println(lease);
//			System.out.println(lease);
//			lease.renew(1*24*3600000);
//			System.out.println(lease);
//			System.out.println(js.read(entry, null, 20000000));
//			Thread.sleep(900000);
		} catch (Exception e) {
			e.printStackTrace();
		}
	}


}
